/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 */

package org.objectweb.dream.dreamannotation.generator.template.binding;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.generator.template.monolog.DreamInitLoggerTemplate;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Cardinality;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.generator.template.binding.BindFcTemplate;
import org.objectweb.fractal.fraclet.annotation.generator.template.binding.ListFcTemplate;
import org.objectweb.fractal.fraclet.annotation.generator.template.monolog.LogTemplate;
import org.objectweb.fractal.fraclet.annotation.generator.template.binding.LookupFcTemplate;
import org.objectweb.fractal.fraclet.annotation.generator.template.binding.NonTypedBindFcTemplate;
import org.objectweb.fractal.fraclet.annotation.generator.template.binding.UnbindFcTemplate;
import org.objectweb.fractal.fraclet.annotation.generator.template.util.NonTypedFractalAttributeException;
import org.objectweb.fractal.fraclet.annotation.generator.template.util.None;

import spoon.reflect.Factory;
import spoon.reflect.code.CtBlock;
import spoon.reflect.code.CtStatementList;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtField;
import spoon.reflect.reference.CtTypeReference;
import spoon.template.Local;
import spoon.template.Parameter;
import spoon.template.Substitution;
import spoon.template.Template;
import spoon.template.TemplateParameterList;

/**
 * A template code to insert the binding controller methods. <br />
 * <br />
 * The only difference with the fraclet implementation is the recursive
 * treatment of Component defined with the <code>@DreamComponent</code> annotation.
 * @author Pierre Garcia <pierre.garcia@inrialpes.fr>
 */
public class DreamRequiresTemplate implements Template {

	@Local
	private static final boolean DEBUG = false;

	@Parameter
	TemplateParameterList<?> _listFcAdds_;

	@Parameter
	TemplateParameterList<?> _unbinds_;

	@Parameter
	TemplateParameterList<?> _lookups_;

	@Parameter
	TemplateParameterList<?> _binds_;

	@Local
	Factory spoonFactory;
	@Local
	CtStatementList<?> listFcAdds;
	@Local
	CtStatementList<?> unbinds;
	@Local
	CtStatementList<?> binds;
	@Local
	CtStatementList<?> lookups;

	/**
	 * Constructor for RequiresTemplate. This constructor is called from the
	 * <code>@Requires</code> Processor
	 * @param spoonFactory
	 *            the factory
	 * @param target
	 *            the target class
	 */
	@Local
	public DreamRequiresTemplate(Factory spoonFactory, CtClass<?> target)
			throws NonTypedFractalAttributeException {
		// 4 sub templates are used, one for each method to insert
		this.spoonFactory = spoonFactory;
		listFcAdds = spoonFactory.Core().createStatementList();
		unbinds = spoonFactory.Core().createStatementList();
		binds = spoonFactory.Core().createStatementList();
		lookups = spoonFactory.Core().createStatementList();
		CtTypeReference<?> parentTarget = (CtTypeReference<?>) target
				.getSuperclass();
		CtClass<?> localTarget = target;
		while (parentTarget != null
				&& ((parentTarget.getDeclaration().getAnnotation(
						FractalComponent.class) != null) || (parentTarget
						.getDeclaration().getAnnotation(DreamComponent.class) != null))) {
			if (DEBUG)
				System.out.println("PROCESSING CLASS "
						+ parentTarget.getQualifiedName());
			parseFields(target, (CtClass<?>) parentTarget.getDeclaration(),
					spoonFactory);
			localTarget = (CtClass<?>) parentTarget.getDeclaration();
			parentTarget = (CtTypeReference<?>) localTarget.getSuperclass();
		}
		if (DEBUG)
			System.out.println("PROCESSING CLASS " + target.getQualifiedName());
		parseFields(target, target, spoonFactory);
		// set the references to the list statements
		_listFcAdds_ = listFcAdds;
		_unbinds_ = unbinds;
		_binds_ = binds;
		_lookups_ = lookups;
	}

	@SuppressWarnings("unchecked")
	@Local
	private void parseFields(CtClass<?> target, CtClass<?> parentTarget,
			Factory spoonFactory) throws NonTypedFractalAttributeException {
		List<CtField<?>> fields = parentTarget.getFields();

		Service serviceAnnotation = null;
		Monolog logAnnotation;

		// iterate over target fields
		for (CtField<?> field : fields) {

			if (DEBUG)
				System.out.println("Field " + field.getSimpleName() + " of "
						+ parentTarget.getQualifiedName());

			// ////////////////////////////////////////////////////
			// /////// Manage the @Service annotation //////////
			// ///////////////////////////////////////////////////

			serviceAnnotation = field.getAnnotation(Service.class);
			if (serviceAnnotation != null) {
				if (target == parentTarget) {
					if (DEBUG)
						System.out.println("Service Annotaton in "
								+ field.getParent(CtClass.class)
										.getQualifiedName());
					treatServiceAnnotation(target, spoonFactory, field);
					Template dilt = new DreamInitLoggerTemplate();
					Substitution.insertAll(target, dilt);
				}
			}

			// ////////////////////////////////////////////////////
			// /////// Manage the @Monolog annotation //////////
			// ///////////////////////////////////////////////////

			logAnnotation = field.getAnnotation(Monolog.class);
			if (logAnnotation != null) {
				if (field.getParent().getAnnotation(DreamComponent.class) != null) {
					System.err
							.println("\n#### ERROR #### \nA field annotated with @Monolog in the class "
									+ field.getParent(CtClass.class)
											.getQualifiedName()
									+ " can't be processed because this class is annotated with @DreamComponent.\n"
									+ "Two solutions depending on your needs \n"
									+ "- Replace @Monolog by @DreamMonolog for a dream management of logger \n"
									+ "- Replace @DreamComponent by @FractalComponent for a default fraclet management of logger \n"
									+ "###############\n");
					System.exit(1);
				} else {
					if (!target.getSuperInterfaces().contains(
							spoonFactory.Type().createReference(
									BindingController.class))) {
						target.getSuperInterfaces().add(
								spoonFactory.Type().createReference(
										BindingController.class));
					}

					Template logT = new LogTemplate(field.getSimpleName());
					CtBlock<?> rcBlock = Substitution.substituteMethodBody(
							target, logT, "bind", new CtTypeReference<?>[] {
									spoonFactory.Type().createReference(
											String.class),
									spoonFactory.Type().createReference(
											Object.class) });
					binds.getStatements().addAll(rcBlock.getStatements());
				}
			}
			// //////////////////////////////////////////////////
			// //////////////////////////////////////////////////

			// get the Requires annotation
			Requires requiresAnnotation = field.getAnnotation(Requires.class);
			// if the field is not annotated continue
			if (requiresAnnotation == null)
				continue;

			if (!target.getSuperInterfaces().contains(
					spoonFactory.Type()
							.createReference(BindingController.class))) {
				target.getSuperInterfaces().add(
						spoonFactory.Type().createReference(
								BindingController.class));
			}

			CtBlock<?> block = null;
			Template t = null;
			// //////////////////////////////////////////////////
			// If the cardinality is singleton
			// //////////////////////////////////////////////////
			if (requiresAnnotation.cardinality().equals(Cardinality.SINGLETON)) {
				// //////////////////////////////////////////////////
				// the listFcTemplate
				// //////////////////////////////////////////////////
				t = new ListFcTemplate(requiresAnnotation.name(), false);
				block = Substitution.substituteMethodBody(target, t,
						"addSingleton", spoonFactory.Type().createReference(
								List.class));
				// insert all the "interfaces.add("myService")"
				listFcAdds.getStatements().add(block.getStatements().get(0));
				// //////////////////////////////////////////////////
				// the UnbindFcTemplate
				// //////////////////////////////////////////////////
				t = new UnbindFcTemplate(requiresAnnotation.name(), field
						.getType(), field.getSimpleName());
				block = Substitution.substituteMethodBody(target, t,
						"unbindSingleton", spoonFactory.Type().createReference(
								String.class));
				unbinds.getStatements().add(block.getStatements().get(0));
				// //////////////////////////////////////////////////
				// the BindFcTemplate
				// //////////////////////////////////////////////////
				t = new BindFcTemplate(requiresAnnotation.name(), field
						.getType(), field.getSimpleName());
				block = Substitution.substituteMethodBody(target, t,
						"bindSingleton", new CtTypeReference<?>[] {
								spoonFactory.Type().createReference(
										String.class),
								spoonFactory.Type().createReference(
										Object.class) });
				binds.getStatements().add(block.getStatements().get(0));
				// //////////////////////////////////////////////////
				// the LookupFcTemplate
				// //////////////////////////////////////////////////
				t = new LookupFcTemplate(requiresAnnotation.name(), field
						.getType(), field.getSimpleName());
				block = Substitution.substituteMethodBody(target, t,
						"lookupSingleton", spoonFactory.Type().createReference(
								String.class));
				lookups.getStatements().add(block.getStatements().get(0));
			}
			// //////////////////////////////////////////////////
			// If the cardinality is collection
			// //////////////////////////////////////////////////
			else if (requiresAnnotation.cardinality().equals(
					Cardinality.COLLECTION)) {

				t = new ListFcTemplate(field.getSimpleName(), true);
				block = Substitution.substituteMethodBody(target, t,
						"addCollection", spoonFactory.Type().createReference(
								List.class));
				// insert all the "interfaces.add("myService")"
				listFcAdds.getStatements().add(block.getStatements().get(0));
				// //////////////////////////////////////////////////
				// the UnbindFcTemplate
				// //////////////////////////////////////////////////
				t = new UnbindFcTemplate(field.getSimpleName());
				block = Substitution.substituteMethodBody(target, t,
						"unbindCollection", spoonFactory.Type()
								.createReference(String.class));
				unbinds.getStatements().add(block.getStatements().get(0));
				// //////////////////////////////////////////////////
				// the BindFcTemplate
				// //////////////////////////////////////////////////
				if (requiresAnnotation.signature().equals(Class.class)
						&& field.getType().getActualTypeArguments().size() < 2)
					throw new NonTypedFractalAttributeException(
							"Fractal attribute "
									+ field.getSimpleName()
									+ " must be typed to generate the appropriate binding controller implementation");
				else if (!requiresAnnotation.signature().equals(None.class)) {
					t = new NonTypedBindFcTemplate(requiresAnnotation.name(),
							field.getSimpleName(), spoonFactory.Type()
									.createReference(
											requiresAnnotation.signature()));
				} else {
					t = new BindFcTemplate(requiresAnnotation.name(), field
							.getSimpleName(), field.getType()
							.getActualTypeArguments().get(1));
				}
				block = Substitution.substituteMethodBody(target, t,
						"bindCollection", new CtTypeReference<?>[] {
								spoonFactory.Type().createReference(
										String.class),
								spoonFactory.Type().createReference(
										Object.class) });
				binds.getStatements().add(block.getStatements().get(0));
				// //////////////////////////////////////////////////
				// the LookupFcTemplate
				// //////////////////////////////////////////////////
				t = new LookupFcTemplate(field.getSimpleName());
				block = Substitution.substituteMethodBody(target, t,
						"lookupCollection", spoonFactory.Type()
								.createReference(String.class));
				lookups.getStatements().add(block.getStatements().get(0));
			}
			// field.getAnnotations().remove(field.getAnnotation(spoonFactory.Type().createReference(Requires.class)));
		}

		// Find the service in super class
		if (target == parentTarget && serviceAnnotation == null) {
			CtField<?> serviceFieldInSuperClass = getServiceAnnotationInSuperClass(target);
			if (serviceFieldInSuperClass != null) {
				if (DEBUG)
					System.out.println("service annotation in superclass of "
							+ target.getQualifiedName());
				treatServiceAnnotation(target, spoonFactory,
						serviceFieldInSuperClass);
			}
		}

	}

	@Local
	private CtField<?> getServiceAnnotationInSuperClass(CtClass<?> target) {
		CtTypeReference<?> superClassDeclaration = target.getSuperclass();
		CtClass<?> superClass;
		while (superClassDeclaration != null) {
			superClass = (CtClass<?>) superClassDeclaration.getDeclaration();
			CtField<?> serviceField = getServiceField(superClass);
			if (serviceField != null) {
				return serviceField;
			}
			superClassDeclaration = superClassDeclaration.getSuperclass();
		}
		return null;
	}

	@Local
	private CtField<?> getServiceField(CtClass<?> superClass) {
		List<CtField<?>> fields = superClass.getFields();
		Service serviceAnnotation;
		for (CtField<?> field : fields) {
			serviceAnnotation = field.getAnnotation(Service.class);
			if (serviceAnnotation != null)
				return field;
		}
		return null;
	}

	@Local
	private void treatServiceAnnotation(CtClass<?> target,
			Factory spoonFactory, CtField<?> field) {
		if (!target.getSuperInterfaces().contains(
				spoonFactory.Type().createReference(BindingController.class))) {
			target.getSuperInterfaces().add(
					spoonFactory.Type()
							.createReference(BindingController.class));
		}

		Template rcT = new DreamBindRcTemplate(field.getSimpleName());
		CtBlock<?> rcBlock = Substitution.substituteMethodBody(target, rcT,
				"bind", new CtTypeReference<?>[] {
						spoonFactory.Type().createReference(String.class),
						spoonFactory.Type().createReference(Object.class) });
		binds.getStatements().add(rcBlock.getStatements().get(0));
	}

	/**
	 * (non-Javadoc)
	 * 
	 * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
	 *      java.lang.Object) Method automatically generated with Spoon
	 *      <http://spoon.gforge.inria.fr>
	 */
	public void bindFc(String clientItfName, Object serverItf)
			throws NoSuchInterfaceException, IllegalBindingException,
			IllegalLifeCycleException {
		_binds_.S();

		throw new NoSuchInterfaceException("Client interface \'"
				+ clientItfName + "\' is undefined.");
	}

	/**
	 * (non-Javadoc)
	 * 
	 * @see org.objectweb.fractal.api.control.BindingController#listFc() Method
	 *      automatically generated with Spoon <http://spoon.gforge.inria.fr>
	 */
	public String[] listFc() {
		List<String> __interfaces__ = new ArrayList<String>();
		_listFcAdds_.S();
		return __interfaces__.toArray(new String[__interfaces__.size()]);
	}

	/**
	 * (non-Javadoc)
	 * 
	 * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
	 *      Method automatically generated with Spoon
	 *      <http://spoon.gforge.inria.fr>
	 */
	public Object lookupFc(String clientItfName)
			throws NoSuchInterfaceException {
		_lookups_.S();
		throw new NoSuchInterfaceException("Client interface \'"
				+ clientItfName + "\' is undefined.");
	}

	/**
	 * (non-Javadoc)
	 * 
	 * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
	 *      Method automatically generated with Spoon
	 *      <http://spoon.gforge.inria.fr>
	 */
	public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
			IllegalBindingException, IllegalLifeCycleException {
		_unbinds_.S();
		throw new NoSuchInterfaceException("Client interface \'"
				+ clientItfName + "\' is undefined.");
	}
}